#!/usr/bin/perl
# -*- cperl -*- # force cperl mode
use strict;
use POSIX qw/WIFEXITED WEXITSTATUS WIFSIGNALED WTERMSIG WIFSTOPPED WSTOPSIG/;
use IO::Handle;
use Data::Dumper;

=comments

This short script will run python against modules specified on the command line,
and print gcc-style error messages instead of python backtrace ones

check.pl [-d <directory>] file1 file2 file3 ...
-d <dir> : strip leading <directory> from the reported filenames (makes output nicer looking)

=cut

# only one option, do it the cheapO way
my $directory_strip = undef;
if ($ARGV[0] =~ /-d/) {
    shift(@ARGV); # toss out flag
    $directory_strip = shift(@ARGV); # take the next arg
}

STDOUT->autoflush(1);

my $errors_found = 0;
foreach my $file (@ARGV) {
    print "$file ... ";

    # pass1 : make sure it parses
    my ($local_errors_cnt, $local_errors_str) = python_load_pass($file);
    $local_errors_cnt += grep_source_pass($file);

    if (!$local_errors_cnt) {
        print "ok\n";
        if (length($local_errors_str)) {
            print $local_errors_str;
        }
    } else {
        print "failure\n";
        print $local_errors_str;
    }

    $errors_found += $local_errors_cnt;
}

exit($errors_found);

sub grep_source_pass {
    my $file = shift;
    my $error_cnt = 0;

    my @module = ();
    {
        open(FH, "<$file") || die("couldn't open '$file'");
        @module = <FH>;
        close(FH);
    }

    my $has_version_str = 0;
    my %imports = ();
    for (my $i = 0; $i < scalar(@module); $i++) {
        my $line = $module[$i];
        my $line_number = $i + 1;

        if ($line =~ /__version__/) {
            $has_version_str = 1;
        }

#        if ($line =~ /from (\w+) import (.*)/) {
#            print "ERROR: from $1 import $2\n";
#            $error_cnt += 1;
#        }

        if ($line =~ /import (.*)/) {
            @imports{split(/\s*,\s*/, $1)} = (1..100); # it will be less than 100 imports
        }

        # split and check all lines like foo.bar()
        if ($line =~ /\w+\.\w+/) {
            # split on whitespace
            my @arr = split(/[^\w.]+/, $line);
            foreach my $guy (@arr) {
                if ($guy =~ /([\w.]+)\.\w+$/) { # bp = Graph.BarPlot(ydata), we are looking at the Graph.BarPlot
                    delete $imports{$1};
                    if ($1 =~ /([\w]+)\./) {
                        delete $imports{$1};
                    }
                }
            }
        }
    }

    if (%imports) { # imports that weren't used
        foreach my $mod (keys(%imports)) {
            print "Module '$mod' imported but not used\n";
# just warn            $error_cnt += 1;
        }
    }

    unless ($has_version_str) {
        print "No __version__ string in file\n";
        $error_cnt += 1;
    }
    return $error_cnt;
}

sub python_load_pass {
    my $file = shift;
    my $error_cnt = 0;

    $ENV{PYTHON_CHECKER} = 1; # tell our Python that it isn't really being run (Page::Paint just returns, for example)
    $file =~ s/\.py$//;
    my $fake_file = <<CONTENTS;
import $file
i = 0
i += 1
CONTENTS

    my $tfile = "/tmp/check_python.$$";
    open(TF,">$tfile") || die("couldn't create tempfile");
    print TF $fake_file;
    close(TF);
    open(FH, "cat $tfile | python -O -OO - 2>&1 |") || die("couldn't open file '$tfile'");

    my $err_output = '';
    while(<FH>) {
        $err_output .= $_;
    }

    close(FH);
    unlink($tfile); # we have to do this AFTER the popen() above is complete and closed!

    if (!WIFEXITED($?)) { # exited abnormally
        die(WEXITSTATUS($?)); # die with the signal from the guy
    }
    return ($error_cnt, $err_output);
}
